import asyncio
import json
import sys
import typing as t
import uuid
from asyncio import Future
from collections import deque
from collections.abc import AsyncGenerator
from dataclasses import is_dataclass, asdict

from ajsonrpc.core import JSONRPC20Response, JSONRPC20Request

from celestia.types import Base64
from .abc import RPCExecutor, Transport, logger

if sys.version_info[:2] <= (3, 10):
    from async_timeout import timeout as asyncio_timeout

    asyncio.timeout = asyncio_timeout

RPC_VALUE_ERRORS = [
    'unmarshaling params',
    'equal to 0',
    'given height is from the future',
    'invalid range',
    'height must be bigger than zero',
    'dial to self attempted',
    'gater disallows connection to peer',
    'notfound desc = delegation with delegator',
    'unknown desc = failed to execute message; message index: 0: invalid shares amount:',
    'cannot redelegate to the same validator',
    'too many unbonding delegation entries for (delegator, validator) tuple',
    'redelegation not found for delegator address',
    'too many redelegation entries for (delegator, src-validator, dst-validator)',
    'datastore: key not found'
]


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if is_dataclass(obj):
            return asdict(obj)
        if isinstance(obj, Base64):
            return str(obj)
        return super().default(obj)


class RPC(RPCExecutor):
    """ RPC encoder / executor / decoder
    """

    def __init__(self, transport: Transport, timeout: float = 180):
        self.timeout = timeout
        self.transport = transport
        self.transport.on_message = self.on_transport_response
        self.transport.on_close = self.on_transport_close
        self._pending = dict()  # type: dict[str, Future]
        self._subscriptions = dict()  # type: dict[str, deque[t.Any]]

    def on_transport_response(self, message: str):
        message = json.loads(message)
        if 'method' in message:
            subscription_id, item = message['params']
            subscription = self._subscriptions.get(subscription_id, None)
            if subscription is not None:
                subscription.append(item)
        else:
            response = JSONRPC20Response(result=False)
            response.body = message
            if future := self._pending.get(response.id):
                if response.error is not None:
                    error_body = getattr(response.error, "body", None)
                    error_message = error_body.get('message', None).lower() if error_body else None
                    if any(keyword in error_message for keyword in RPC_VALUE_ERRORS):
                        future.set_exception(ValueError(error_message))
                    elif error_message is None or error_body is None:
                        future.set_exception(ConnectionError("RPC failed; undefined error"))
                    else:
                        future.set_exception(ConnectionError(f"RPC failed; {error_message}", response.error))
                else:
                    future.set_result(response.result)
            else:
                logger.warning("Received message with unexpected ID.")

    def on_transport_close(self, exc: Exception = None):
        if exc:
            e = ConnectionError(f"RPC failed; transport closed by error {exc}")
            exc = e.with_traceback(exc.__traceback__)

        for future in self._pending.values():
            if exc:
                future.set_exception(exc)
            else:
                future.set_exception(ConnectionError("RPC failed; transport closed"))

        for id in tuple(self._subscriptions.keys()):
            subscription = self._subscriptions.pop(id)
            if exc:
                subscription.append(exc)

    async def call(self, method: str, params: tuple[t.Any, ...] = None,
                   deserializer: t.Callable[[t.Any], t.Any] = None) -> t.Any | None:
        params = params or ()
        deserializer = deserializer or (lambda a: a)
        id = str(uuid.uuid4())
        request = JSONRPC20Request(method, params, id)
        await self.transport.send(json.dumps(request.body, cls=JSONEncoder))
        future = self._pending[id] = Future()
        future.add_done_callback(lambda _: self._pending.pop(id, None))
        async with asyncio.timeout(self.timeout):
            result = await future
            return deserializer(result)

    async def iter(self, method: str, params: tuple[t.Any, ...] = None,
                   deserializer: t.Callable[[t.Any], t.Any] = None) -> AsyncGenerator[t.Any]:
        deserializer = deserializer or (lambda a: a)
        subscription_id = await self.call(method, params)
        try:
            subscription = self._subscriptions[subscription_id] = deque()
            while subscription_id in self._subscriptions:
                if len(subscription):
                    item = subscription.popleft()
                    if isinstance(item, Exception):
                        raise item
                    yield deserializer(item)
                else:
                    await asyncio.sleep(0.1)
        finally:
            del self._subscriptions[subscription_id]



import asyncio
from celestia.node_api import Client

async def test_graceful_close_subscribing(node_provider):
    result = []
    bridge, auth_token = await node_provider('bridge-0')
    client = Client(port=bridge.port['26658/tcp'])
    height_deserializer=lambda data: int(data['header']['height'])

    async with client.connect(auth_token) as api:
        start_height = await api.header.local_head(deserializer=height_deserializer)

    async def subscriber():
        try:
            async with client.connect(auth_token) as api:
                async with asyncio.timeout(30):
                    async for height in api.header.subscribe(deserializer=height_deserializer):
                        result.append(height)
        except asyncio.CancelledError:
            pass

    subscriber_task = asyncio.create_task(subscriber())

    async with asyncio.timeout(30):
        while True:
            if len(result) == 3:
                subscriber_task.cancel()
                await subscriber_task  # This is important!
                break
            await asyncio.sleep(0.1)

    assert start_height + 3 == result[-1]
    assert len(result) == 3




import asyncio
from contextlib import asynccontextmanager, AbstractAsyncContextManager
from urllib.parse import urlparse

from websockets.asyncio.client import connect, ClientConnection

from .abc import Transport as AbcTransport
from .executor import RPC


class Client:
    """ WS JSON-RPC Client
    """

    def __init__(self, url: str = None, /, *, auth_token: str =None,
                 host: str = 'localhost', port: int = 26658):
        url = self._make_url(url, host=host, port=port)
        self.options = dict(url=url, auth_token=auth_token)

    def _make_url(self, url: str = None, /, *,
                  host: str = 'localhost', port: int = 26658, protocol: str = 'ws') -> str:
        if url is None:
            url = f'{protocol}://{host}:{port}'
        pr = urlparse(url or "ws://localhost:26658")
        if pr.scheme not in ("ws", "wss"):
            raise ValueError("Unsupported URL scheme, must be ws or wss")
        return f'{pr.scheme}://{pr.hostname}:{pr.port or port}'

    def connect(self, auth_token: str = None, /, *,
                response_timeout: float = 180) -> AbstractAsyncContextManager[RPC]:
        headers = []
        url = self.options['url']
        if auth_token := auth_token or self.options['auth_token']:
            headers.append(('Authorization', f'Bearer {auth_token}'))

        async def listener(connection: ClientConnection, handlers: AbcTransport):
            try:
                async for message in connection:
                    handlers.on_message(message)
            except asyncio.CancelledError:
                handlers.on_close(None)
            except Exception as exc:
                handlers.on_close(exc)

        @asynccontextmanager
        async def connect_context():
            async with connect(url, additional_headers=headers) as connection:
                class Transport(AbcTransport):

                    async def send(self, message: str):
                        await connection.send(message)

                transport = Transport()
                rpc = RPC(transport, response_timeout)
                self._listener_task = asyncio.create_task(listener(connection, transport))
                yield rpc
                self._listener_task.cancel()
                await self._listener_task

        return connect_context()



import pytest
from pydantic import BaseModel

from celestia.node_api import Client


@pytest.mark.asyncio
async def test_deserializers(node_provider):
    bridge, auth_token = await node_provider('bridge-0')
    client = Client(port=bridge.port['26658/tcp'])

    class CustomBalanceModel(BaseModel):
        amount: int
        denom: str

    async with client.connect(auth_token) as api:
        last_height = await api.header.local_head(deserializer=lambda data: int(data['header']['height']))
        assert isinstance(last_height, int)

        balance = await api.state.balance(deserializer=CustomBalanceModel.model_validate)
        assert isinstance(balance.amount, int)
        assert balance.amount

        with pytest.raises(TypeError):
            await api.state.balance(CustomBalanceModel.model_validate)



import logging
import typing as t
from abc import ABC, abstractmethod
from collections.abc import AsyncGenerator

logger = logging.Logger('WS JSON-RPC')

class Transport(ABC):
    on_message: t.Callable[[bytes | str], None]
    on_close: t.Callable[[Exception | None], None]

    @abstractmethod
    async def send(self, message: str) -> None:
        """ Send a message to the connection. """


class RPCExecutor(ABC):

    @abstractmethod
    async def call(self, method: str, params: tuple[t.Any, ...] = None,
                   deserializer: t.Callable[[t.Any], t.Any] = None) -> t.Any | None:
        """ This method must implement calling an RPC method and returning the result.
        """

    @abstractmethod
    async def iter(self, method: str, params: tuple[t.Any, ...] = None,
                   deserializer: t.Callable[[t.Any], t.Any] = None) -> AsyncGenerator[t.Any]:
        """ This method must implement creating an `RPC` subscription and returning
        an asynchronous iterator that returns the incoming subscription results.
        """


class Wrapper:
    def __init__(self, rpc: RPCExecutor):
        self._rpc = rpc



from contextlib import asynccontextmanager, AbstractAsyncContextManager

from . import rpc
from .blob import BlobClient
from .das import DasClient
from .fraud import FraudClient
from .header import HeaderClient
from .p2p import P2PClient
from .rpc import RPC
from .share import ShareClient
from .state import StateClient


class NodeAPI:
    """ Celestia node API """

    def __init__(self, rpc: RPC):
        self._rpc = rpc

    @property
    def state(self):
        return StateClient(self._rpc)

    @property
    def blob(self):
        return BlobClient(self._rpc)

    @property
    def header(self):
        return HeaderClient(self._rpc)

    @property
    def p2p(self):
        return P2PClient(self._rpc)

    @property
    def das(self):
        return DasClient(self._rpc)

    @property
    def fraud(self):
        return FraudClient(self._rpc)

    @property
    def share(self):
        return ShareClient(self._rpc)


class Client(rpc.Client):
    """ Celestia Node API client
    """

    def connect(self, auth_token: str = None, /, response_timeout: float = 180) -> AbstractAsyncContextManager[NodeAPI]:
        """ Creates and return connection context manager. """

        @asynccontextmanager
        async def connect_context():
            async with super(Client, self).connect(auth_token, response_timeout=response_timeout) as rpc:
                yield NodeAPI(rpc)

        return connect_context()
